#include <alloc.h>
#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <FCNTL.H>

  /* Tipos de variables */
#define ENTRADA 1
#define SALIDA  2

  /* tipos de conjuntos */	
#define TRIANGULAR 1
#define TRAPEZOIDAL 2

  /* Operadores posibles*/
#define AND 1
#define OR  2
#define NOT 3
#define EQ  4
#define PER 5

//*************************************************************
//         Estructuras de datos

typedef struct per
	{
	 double grado;
	 struct per *next;
	} lista_per ;  /* Estrcutura para mantener una lista con todos los grados
							de pertenencia a un cnj de una vble de salida */

typedef struct cnj
	{
	 char nombre[255] ; /* Identificador del cnj */
	 char tipo;         /* Forma del cnj Fuzzy ( tiangular o trapezaoidal ) */
	 double left ;      /* Lmite inferior del cnj */
	 double right;      /* Lmite superior del cnj */
	 double top_left ;  /* Vertice superior izquierdo del cnj */
	 double top_right ; /* Vertice superior derecho del cnj */
	 double grado_pertenencia ; /* Guarda en un momento dado el grado de 
											 pertenencias de un variable
											 al cnj, es decir cuando para un valor 
											 determinado de la vble se concule su grado de
											 pertenencia al cnj se guardar en este campo
											 y a partir de el se realizarn las inferencias
										 */
	 lista_per *ptr_lista;  /* lista para mantener los grados de pertenencia
										de una vble de salida */
	 struct cnj *next, *last; /* Punteros para poder mantener una lista 
									 doblemente enlazada de los cnj de una vble */
	 double area;   /* Area total del conjunto */
	 double px ;    /* Resultado de INTEGRAL(P(x) * x dx) */

	} cnj_fuzzy;


/* Estructura para mantener la informacin referente a un vble en el sistema*/
typedef struct variable
	{
	 char nombre[255]; /* 255 caracter para guardar un nombre significativo
						  para la vble */
	 char tipo;  /* Determina si la vble es de salida o de entrada */
	 double min;
	 double max ; /* Rango de valores que puede tomar la vble */
	 cnj_fuzzy *conjuntos; /* Puntero para direccionar el comienxo de una
										lista enlazada con todos los conjuntos.*/
	 struct variable *next; /* Para mantener una lista con las variable del 
								sistema */
	 char fichero[255]; /* para almacenar el nombre del fichero de donde
									  se tomarn los datos */	
	} vble;



/* Cada regla se guardar en una estructura de rbol */
typedef struct reg
	{
	 char operador; /* Indica en tipo de operacin que se debe ejecutar en
										 en cada nodo */
	 union nodo
		{
		 struct nohoja /* Estructura en caso de que no se trate de un nodo hoja */
			{
			 struct reg *izq,*der ;  /* punteros para mantener una estructura de arbol */
			} nohoja;
		 struct hoja  /* Estructura en caso de que se trate de un nodo hoja */
			{
			 vble *variable; /* puntero a la vble en cuestin*/
			 cnj_fuzzy *cnj;  /* puntero al cnj que debe pertenecer */
			} hoja ;
		} nodo;
	}  regla;

/* Estructura para mantener una lista de reglas en memoria */
typedef struct lis_reg
	{
	 regla *reg;
	 struct lis_reg *next;
	} lis_reg;



//***************************************************************
//          DEFINICION DE PROTOTIPOS



double pertenencia(cnj_fuzzy cnj,double value);
void valor_entrada(vble v, double valor );
double calculo_regla(regla r);
void init_cnj(cnj_fuzzy &cnj);
double media_producto_suma(vble v);
void abrir(char *nombre);  // funcion para leer un fichero de reglas
void motor_inferencia(double *vbles_entrada, double *vbles_salida);

// vble glovales para mantener el sistema de inferencia fuzzy

vble *vble_entrada = NULL;
vble *vble_salida = NULL;
lis_reg *lista_reglas = NULL;



extern unsigned _stklen = 20000U;




/* 
	FUNCION : pertenencia;
	OBJETIVO : Dato un conjunto fuzzy y un valor debe hallar el valor de la 
		funcin de asociacin del valor al conjunto y devolverla
	PARAMETROS:
		cnj   --> conjunto fuzzy
		value --> valor determinado de la  vble en un determinado momento;
*/

double pertenencia(cnj_fuzzy cnj,double value)
	{
	 double BaseL, BaseR, RatioL = 0, RatioR = 0;

	 if ( (value < cnj.left) || ( value > cnj.right))
		return 0;  /* Si el valor esta fuera de los lmites del cnj */

	 BaseL = cnj.top_left - cnj.left;
	 BaseR = cnj.right - cnj.top_right;
	 if ( BaseL > 0 ) 
		RatioL = ( 100 / BaseL);
	 if ( BaseR > 0 ) 
		RatioR = ( 100 / BaseR); /* se calcula la pentiente de los borde de los cnj*/

	 if ( cnj.tipo == TRIANGULAR )
		{
		 if ( value < cnj.top_left)
			return (( value - cnj.left) * RatioL);
		 else
			if ( value > cnj.top_left)
				return (( cnj.right - value) * RatioR);
			else
				return 100;
		}
	 else  // trapeziodal
		{
		 if ( (value >= cnj.top_left ) && ( value <= cnj.top_right))
			return (100);
		 if ( value <= cnj.top_left)
			return (( value - cnj.left) * RatioL);
		 else
			return (( cnj.right - value) * RatioR);

		}
	}




/*
	FUNCION : valor_entrada
	OBJETIVO : Funcin que dada una variable y un valor determinado, halla la 
		funcin de pertenencia a todos sus cnj.
	FUNCIONES USADAS:
		pertenencia.
	PARAMETROS:
		vble  --> variable fuzzy
		value --> valor determinado de la  vble en un determinado momento;
*/

void valor_entrada(vble v, double value )
	{
	 cnj_fuzzy *p;

	 p = v.conjuntos;
	 while ( p != NULL)
		{
		 p->grado_pertenencia = pertenencia(*p,value);
		 p = p->next;
		}
	}



double minimo(double a, double b)
	{
	 return ( (a < b) ? a : b);
	}

double maximo(double a, double b)
	{
	 return ( (a > b) ? a : b);
	}

/* 
	FUNCION : calculo_regla
	OBJETIVO : Realizar el cmputo de una regla. Para ello se recorre la 
		estructura de rbol donde se almacena la regla realizando en cada
		caso la operacin que se indique en el nodo.
		El valor que resulta se almacenara el en cnj correspondiente de 
		la variable de salida
	PARAMETROS:
		r --> regla a evaluar;
*/

double calculo_regla(regla r)
	{
	 regla *p;
	 double result;
	 cnj_fuzzy *cnj;
	 lista_per *lp;
	 lista_per *q;
	 
	 switch (r.operador)
		{
		 case EQ:
					p = r.nodo.nohoja.der; /* La parte derache tendra la variable de salida */	
					/* El grado de pertenencia de la vble de salida ser la 
						evaluacin del lado izquierdo del arbol */
					result = calculo_regla(*r.nodo.nohoja.izq);
					lp = (lista_per *)farmalloc(sizeof(lista_per));
					if ( lp == NULL)
						{
						 printf("\nERROR: Memoria insuficiente");
						 exit(-1);
						}
					lp->grado = result;
					lp->next = NULL;
					cnj = p->nodo.hoja.cnj;
					if ( cnj->ptr_lista == NULL)
						{
						 cnj->ptr_lista = lp;
						}
					else
						{
						 q = cnj->ptr_lista;
						 while ( q->next != NULL)
							q = q->next;
						 q->next = lp;
						}
					return 0;
		 case AND:
					return ( minimo(calculo_regla(*r.nodo.nohoja.izq),calculo_regla(*r.nodo.nohoja.der)));
		 case OR :
					return ( maximo(calculo_regla(*r.nodo.nohoja.izq),calculo_regla(*r.nodo.nohoja.der)));
		 case NOT:
					return ( (double)100 - calculo_regla(*r.nodo.nohoja.izq));
		 case PER: /* se trata de un nodo hoja y se devuelve el valor de
						  pertenencia de la variable al cnj
					  */
					return ( r.nodo.hoja.cnj->grado_pertenencia);
		}
	 return 0;
	}

					
					


/* 
	FUNCION : init_cnj
	OBJETIVO : Calcular el area del cnj y el area debajo de la curva
		P(x)*x Para calcular luego media del producto y la suma.
	PARAMETROS:
		cnj --> conjunto a inicializar;
*/
void init_cnj(cnj_fuzzy &cnj)
	{
	 double BaseL, BaseR, BaseC;
	 double temp;
	 BaseL = cnj.top_left - cnj.left;
	 BaseR = cnj.right - cnj.top_right;

	 if ( cnj.tipo == TRIANGULAR)
		{
		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left - cnj.left * cnj.left) / (double)2;
			 temp -= cnj.left * BaseL;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += BaseR;
			 temp  -= ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / ((double)2 * BaseR);
			 temp += cnj.top_right;
			}
		 cnj.area= (double)100 * temp;

		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left * cnj.top_left - cnj.left * cnj.left * cnj.left) / (double)3;
			 temp -= (cnj.left * (cnj.top_left * cnj.top_left - cnj.left * cnj.left)) / (double)2;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / (double)2;
			 temp  -= ( cnj.right * cnj.right * cnj.right - cnj.top_right * cnj.top_right * cnj.top_right ) / ((double)3 * BaseR);
			 temp += ( cnj.top_right * ( cnj.right * cnj.right - cnj.top_right * cnj.top_right)) / ( (double)2 *BaseR);
			}
		 cnj.px = (double)100 * temp;
		 /* Responde a la frmula

					  100
				px = ---(BaseL^2 + BaseR^2)
						3
		 */
		}
	 else
		{ /* tipo TRAPEZOIDAL */
		 BaseC = cnj.top_right - cnj.top_left;
		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left - cnj.left * cnj.left) / (double)2;
			 temp -= cnj.left * BaseL;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += BaseR;
			 temp  -= ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / ((double)2 * BaseR);
			 temp += cnj.top_right;
			}
		 temp += BaseC;
		 cnj.px = (double)100 * temp;
		 cnj.area =(( 2 * BaseC + BaseL + BaseR) * (double)100) / (double)2;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left * cnj.top_left - cnj.left * cnj.left * cnj.left) / (double)3;
			 temp -= (cnj.left * (cnj.top_left * cnj.top_left - cnj.left * cnj.left)) / (double)2;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / (double)2;
			 temp  -= ( cnj.right * cnj.right * cnj.right - cnj.top_right * cnj.top_right * cnj.top_right ) / ((double)3 * BaseR);
			 temp += ( cnj.top_right * ( cnj.right * cnj.right - cnj.top_right * cnj.top_right)) / ( (double)2 *BaseR);
			}
		 if ( BaseC != 0)
			 temp += (cnj.top_right * cnj.top_right - cnj.top_left * cnj.top_left) / (double) 2;
		 cnj.px = (double)100 * temp;
		 /* Responde a la frmula

							 BaseL^2 + BaseR^2        BaseC^2
			 px= 100 ( --------------------- + -------------- )
									  3		              2
		 */
		}
	}



/*
	FUNCION : media_producto_suma
	OBJETIVO : Funcion que calcula en valor de una vble de salida aplicando
		el mtodo de la media del producto y la suma.
	PARAMETROS:
		cnj --> conjunto a inicializar;
*/

double media_producto_suma(vble v)
	{
	 cnj_fuzzy *p;
	 lista_per *q,*r;
	 double denominador = 0; 
	 double numerador= 0; 
		  /* de la exprecin de la media de la suma de los productos */
			
	 p = v.conjuntos;
	 while ( p != NULL)
		{
		 q = p->ptr_lista;
		 while ( q != NULL)
			{
			 numerador += q->grado * p->px;
			 denominador +=  q->grado * p->area;
			 r = q->next;
			 farfree(q);
			 q = r;
			}
		 p->ptr_lista = NULL;
		 p = p->next;
		}
	 if ( (denominador == 0) || (numerador == 0))
		 return 0;
	 return ( numerador / denominador);
	}


void motor_inferencia(double *vbles_entrada, double *vbles_salida)
   {
	 extern vble *vble_entrada;
	 extern vble *vble_salida;
	 extern lis_reg *lista_reglas;
	 int n_entradas ;
	 int n_salidas  ;
	 int i;
	 vble *p;
    lis_reg *l;


	 // contar variables de entrada

	 p = vble_entrada;
	 n_entradas= 0;
	 while ( p != NULL)
		{
		 n_entradas ++;
		 p = p->next;
		}

	 if ( vble_entrada == 0 )
		{
		 printf("\nNo existen variables de entrada definidas en este sistema\n");
		 return;
		}
	 // contar variables de salida

	 p = vble_salida;
	 n_salidas= 0;
	 while ( p != NULL)
		{
		 n_salidas ++;
		 p = p->next;
		}

	 if ( vble_salida == 0 )
		{
		 printf("\nNo existen variables de salida definidas en este sistema\n");
		 return;
		}

	 p = vble_entrada;
	 for ( i =0; i < n_entradas; i ++, p = p->next )
		{
		 valor_entrada(*p,vbles_entrada[i]);
		}
    l = lista_reglas;
    while(l != NULL)
		{
		 calculo_regla(*(l->reg));
		 l = l->next;
		}
    p = vble_salida;
	 for ( i = 0; i < n_salidas; i ++, p = p->next )
		{
		 vbles_salida[i] = media_producto_suma(*p);
		}
  }





// funcion que lee una reglas del fichero de entrada
regla *leer_regla(int fd)
	{
	 extern unsigned char far *buf1;
	 extern unsigned char far *buf2;
	 extern unsigned char far *buf3;
	 extern unsigned char far *buf4;

	 regla *r;

	 r = (regla *) farmalloc(sizeof(regla));
	 if ( r == NULL)
		{
		 printf("ERROR: memoria insufiente");
		 exit(-1);
		}


	 read(fd, &r->operador,1);

	 switch (r->operador)
		{
		 case EQ:
		 case AND:
		 case OR :
					r->nodo.nohoja.izq = leer_regla(fd);
					r->nodo.nohoja.der = leer_regla(fd);
					break;
		 case NOT:
					r->nodo.nohoja.izq = leer_regla(fd);
					break;
		 case PER:
					{
					 int nv,nc;
					 int i;
					 vble *v;
					 cnj_fuzzy *c;
					 extern vble *vble_entrada;
					 extern vble *vble_salida;


					 read(fd,&nv,sizeof(int));

					 v = vble_entrada;
					 i = 0;
					 while (( v != NULL) && ( i < nv))
						{
						 i ++;
						 if ( i == nv)
							break;
						 v = v->next;
						}
					 if ( v == NULL)
						{
						 v = vble_salida;
						 while (( v != NULL) && ( i < nv))
							{
							 i++;
							 if ( i == nv)
								break;
							 v = v->next;
							}

						}
					 r->nodo.hoja.variable = v;


					 read(fd,&nc,sizeof(int));
					 c = v->conjuntos;
					 i = 0;
					 while((c != NULL) && ( i < nc))
						{
						 i ++;
						 if ( i == nc)
							break;
						 c = c->next;
						}
					 r->nodo.hoja.cnj = c;
					 break;
					}
		}
	 return r;
	}


// funcion para leer datos de un fichero de entrada

void abrir(char *nombre)
	{
	 extern unsigned char far *buf1;
	 extern unsigned char far *buf2;
	 extern unsigned char far *buf3;
	 extern unsigned char far *buf4;

	 extern vble *vble_entrada;
	 extern vble *vble_salida;
	 extern lis_reg *lista_reglas;
	 int n_entradas= 0 ;
	 int n_salidas= 0 ;
	 int i;
	 vble *p;
	 int fd;
	 lis_reg *l,*anl;
	 int n_reglas = 0;
	 vble *v,*anv;



	fd = open(nombre,O_BINARY | O_RDONLY);
	if ( fd == -1)
		{
		 printf("\nError intentando abrir el fichero\n");
		 return;
		}

	read(fd,&n_entradas,sizeof(int));

	for ( i = 0; i < n_entradas; i ++)
	  {
		int n_cnj;
		cnj_fuzzy *c,*anc;
		n_cnj = 0;
		int j;

		v = (vble *) farmalloc(sizeof(vble));
		if ( v == NULL)
		  {
			printf("ERROR: memoria insufiente");
			exit(-1);
		  }

		read(fd,v->nombre,255);
		read(fd,&(v->tipo),1);
		read(fd,&(v->min),sizeof(double));
		read(fd,&(v->max),sizeof(double));
		read(fd,v->fichero,255);

		read(fd,&n_cnj,sizeof(int));
		for (j = 0; j < n_cnj; j ++)
		  {
			c = (cnj_fuzzy *) farmalloc(sizeof(cnj_fuzzy));
			if ( c == NULL)
			  {
				printf("ERROR: memoria insufiente");
				exit(-1);
			  }
			read(fd,c->nombre,255);
			read(fd,&c->tipo,1);
			read(fd,&c->left,sizeof(double));
			read(fd,&c->top_left,sizeof(double));
			read(fd,&c->top_right,sizeof(double));
			read(fd,&c->right,sizeof(double));
			c->ptr_lista = NULL;
			c->next = NULL;
			c->last = NULL;

			if ( j == 0)
			  {
				v->conjuntos = c;
				c->next = NULL;
				c->last = NULL;
            anc = c;
			  }
			else
			  {
				anc->next = c;
				c->last = anc;
				c->next = NULL;
				anc = c;
			  }
		  }
		if ( i == 0)
		  {
			vble_entrada = v;
			v->next = NULL;
			anv = v;
		  }
		else
		  {
			anv->next = v;
			v->next = NULL;
			anv = v;
		  }

	  }

	read(fd,&n_salidas,sizeof(int));

	for ( i = 0; i < n_salidas; i ++)
	  {
		int n_cnj;
		cnj_fuzzy *c,*anc;
		n_cnj = 0;
		int j;

		v = (vble *) farmalloc(sizeof(vble));
		if ( v == NULL)
		  {
			printf("ERROR: memoria insufiente");
			exit(-1);
		  }

		read(fd,v->nombre,255);
		read(fd,&(v->tipo),1);
		read(fd,&(v->min),sizeof(double));
		read(fd,&(v->max),sizeof(double));
		read(fd,v->fichero,255);

		read(fd,&n_cnj,sizeof(int));
		for (j = 0; j < n_cnj; j ++)
		  {
			c = (cnj_fuzzy *) farmalloc(sizeof(cnj_fuzzy));
			if ( c == NULL)
			  {
				printf("ERROR: memoria insufiente");
				exit(-1);
			  }
			read(fd,c->nombre,255);
			read(fd,&c->tipo,1);
			read(fd,&c->left,sizeof(double));
			read(fd,&c->top_left,sizeof(double));
			read(fd,&c->top_right,sizeof(double));
			read(fd,&c->right,sizeof(double));
			c->ptr_lista = NULL;
			c->next = NULL;
			c->last = NULL;

			if ( j == 0)
			  {
				v->conjuntos = c;
				c->next = NULL;
				c->last = NULL;
				anc = c;
			  }
			else
			  {
				anc->next = c;
				c->next = NULL;
				c->last = anc;
				anc = c;
			  }
		  }
		if ( i == 0)
		  {
			vble_salida = v;
			v->next = NULL;
         anv = v;
		  }
		else
		  {
			anv->next = v;
			v->next = NULL;
			anv = v;
		  }

	  }


	read(fd,&n_reglas,sizeof(int));
	for ( i  = 0; i < n_reglas; i ++)
		{
		 l = (lis_reg *) farmalloc(sizeof(lis_reg));
		 if ( l == NULL)
			{
			 printf("ERROR: memoria insufiente");
			 exit(-1);
			}
		 l->reg = leer_regla(fd);
		 if ( i == 0)
			{
			 lista_reglas = l;
			 l->next = NULL;
          anl  = l;
			}
		 else
			{
			 anl->next = l;
			 l->next = NULL;
			 anl = l;
			}
		}

	 close ( fd);
    // inicializar los conjuntos

	 p = vble_entrada;
	 while ( p != NULL)
		{
		 cnj_fuzzy *c;
		 c = p->conjuntos;
		 while ( c != NULL)
			{
			 init_cnj(*c);
			 c = c->next;
			}
		 p = p->next;
		}

	 p = vble_salida;
	 while ( p != NULL)
		{
		 cnj_fuzzy *c;
		 c = p->conjuntos;
		 while ( c != NULL)
			{
			 init_cnj(*c);
			 c = c->next;
			}
		 p = p->next;
		}


	}
